import logging

logger = logging.getLogger("sbio")

# check if logger has been initialized
if not logger.hasHandlers():
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(name)s - %(levelname)s - %(message)s", datefmt="%H:%M:%S"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)



from io import BytesIO

from file_process.csv.csv_processor import CSVFileProcessor
from file_process.exceptions import WrongExtension
from file_process.h5ad.h5ad_processor import H5ADFileProcessor


class FileProcessFactory:  # pylint: disable=too-few-public-methods
    EXTENSIONS_MAP = {
        '.h5ad': H5ADFileProcessor,
        '.csv': CSVFileProcessor
    }

    @classmethod
    def get(cls, filename: str, file: BytesIO, **kwargs):
        for extension, processor_class in cls.EXTENSIONS_MAP.items():
            if filename.endswith(extension):
                return processor_class(file, **kwargs)
        raise WrongExtension

    @classmethod
    def validate_extension(cls, filename: str):
        if not filename:
            raise WrongExtension
        for extension in cls.EXTENSIONS_MAP:
            if filename.endswith(extension):
                return
        raise WrongExtension



from abc import ABC, abstractmethod
from io import BytesIO
from typing import Optional, List


class FileProcessorBase(ABC):
    @abstractmethod
    def validate(self, model_metadata_file: Optional[BytesIO] = None, validation_rules: Optional[dict] = None):
        pass

    @abstractmethod
    def get_preview(self) -> (List[str], List[dict], List[dict]):
        pass



from typing import Optional

import numpy as np
from anndata import AnnData
from scanpy.get import _get_obs_rep
from scipy.sparse import issparse

from file_process.exceptions import NoColumnsError, ModelFileValidationVariablesError, NoXExpression, \
    DataIsNormalized, DataIsNotFinite
from file_process.h5ad.schemas import SbioModelDataForH5ad
from file_process.logger import logger


class H5ADValidator:
    def __init__(self, adata: AnnData, model_data: Optional[SbioModelDataForH5ad] = None, enable_warnings: bool = True):
        self.adata = adata
        self.model_data = model_data
        self.enable_warnings = enable_warnings

    def __call__(self):
        self.validate()
        self.model_file_validation()

    def validate(self):
        self._validate_target_names_present()
        self._check_x()
        self._check_not_normed()
        self._check_finite()

        if self.enable_warnings:
            warnings = []
            warnings += self._get_structure_warnings()
            warnings += self._get_encoding_version_warnings()
            if warnings:
                logger.info("Warnings: %s", warnings)

    def model_file_validation(self):
        if not self.model_data:
            return
        dataset_vars = set(self.adata.var.index)
        difference = self.model_data.var_names - dataset_vars
        if difference:
            raise ModelFileValidationVariablesError(difference)

    def _validate_target_names_present(self):
        target_names = list(self.adata.obs.columns)
        if not target_names:
            raise NoColumnsError

    def _check_x(self):
        if not hasattr(self.adata, 'X') or self.adata.X is None:
            raise NoXExpression

    def _get_structure_warnings(self):
        warnings = []
        if not hasattr(self.adata, 'obs'):
            warnings.append('The h5ad artifact does not contain observation information ".obs".')

        if not hasattr(self.adata, 'var'):
            warnings.append('The h5ad artifact does not contain variable information ".var".')

        if not hasattr(self.adata, 'obsm'):
            warnings.append('The h5ad artifact does not contain experiment design information ".obsm".')

        if not hasattr(self.adata, 'uns'):
            warnings.append('The h5ad artifact does not contain schema information ".uns".')
        return warnings

    def _check_not_normed(self, obs_key: Optional[str] = None):
        data = _get_obs_rep(self.adata, layer=obs_key)
        if issparse(data):
            diff_sum = len(np.unique(np.round(data.A.sum(axis=1), 0), axis=0))
        else:
            diff_sum = len(np.unique(np.round(data.sum(axis=1), 0), axis=0))
        if diff_sum <= 2:
            raise DataIsNormalized

    def _check_finite(self, obs_key: Optional[str] = None):
        data = _get_obs_rep(self.adata, layer=obs_key)
        if issparse(data):
            is_finite = np.isfinite(data.A).all()
        else:
            is_finite = np.isfinite(data).all()
        if not is_finite:
            raise DataIsNotFinite

    def _get_encoding_version_warnings(self):
        # TODO check
        return ["The h5ad artifact was generated with an AnnData version different from 0.8.0."]



from abc import ABC
from typing import Set


class BaseError(ABC, Exception):
    pass


class NotAllTargetsError(BaseError):
    def __init__(self, targets_missing: Set[str]):
        self.message = f'Validation check failed: new data does not contain all targets required by model. ' \
                       f'Missing targets: {targets_missing}'


class NotSomeTargetsError(BaseError):
    def __init__(self, model_targets: Set[str]):
        self.message = f'Validation check failed: new data does not contain any targets required by model. ' \
                       f'List of targets in a model: {model_targets}'


class ModelFileValidationVariablesError(BaseError):
    def __init__(self, variables_missing: Set[str]):
        self.message = f'Validation check failed: new data does not contain all variables (columns) required by ' \
                       f'model. Missing variables: {variables_missing}'


class WrongExtension(BaseError):
    message = 'Cannot process file: wrong extension.'


class DelimiterError(BaseError):
    message = 'Parsing error: try changing delimiter.'


class NoColumnsError(BaseError):
    message = 'No columns in file.'


class CustomValidationException(BaseError):
    def __init__(self, message: str):
        self.message = message


class NoXExpression(BaseError):
    message = 'The h5ad artifact does not contain expression data ".X".'


class DataIsNormalized(BaseError):
    message = 'Data should not be normalized.'


class DataIsNotFinite(BaseError):
    message = 'Data must be finite.'



import json
from io import BytesIO
from typing import List

import pandas as pd

NAME_STR = 'name'
ALLOWED_TYPES_STR = 'allowedTypes'
ALLOWED_VALUES_STR = 'allowedValues'
REQUIRED_STR = 'required'
MIN_STR = 'min'
MAX_STR = 'max'
PRESERVE_ORDER_STR = 'preserveOrder'
COLUMN_NAMES_REQUIRED = 'columnNamesRequired'
ALLOW_OTHER_COLUMNS_STR = 'allowOtherColumns'
COLUMNS_LIST_STR = 'columnsList'


class ValidationRuleError:
    default_path = 'columns'

    def __init__(self, keys: List[str], message: str):
        self.keys = [f'{self.default_path}.{key}' for key in keys]
        self.message = message

    def to_dict(self):
        return self.__dict__


class ColumnValidationRule:

    def __init__(self, validation_rules: dict):
        self._validation_rules = validation_rules
        self.name = validation_rules.get(NAME_STR)
        self.allowed_types = validation_rules.get(ALLOWED_TYPES_STR)
        self.required = validation_rules.get(REQUIRED_STR, True)
        self.allow_missings = validation_rules.get('allowMissings', True)
        self.allow_duplicates = validation_rules.get('allowDuplicates', True)
        self.min = validation_rules.get(MIN_STR)
        self.max = validation_rules.get(MAX_STR)
        self.allowed_values = validation_rules.get(ALLOWED_VALUES_STR)

    def validate_self(self, index: int):  # pylint: disable=too-many-branches
        errors = []
        for key in self._validation_rules.keys():
            if key not in [NAME_STR, ALLOWED_TYPES_STR, REQUIRED_STR, 'allowMissings', 'allowDuplicates', MIN_STR,
                           MAX_STR, ALLOWED_VALUES_STR]:
                errors.append(ValidationRuleError([f'{COLUMNS_LIST_STR}[{index}].{key}'], 'Invalid key'))
        if not self.name:
            errors.append(ValidationRuleError(
                [f'{COLUMNS_LIST_STR}[{index}].{NAME_STR}'],
                'Missing field'
            ))
        if self.allowed_types and len(self.allowed_types) > 1:
            errors.append(ValidationRuleError(
                [f'{COLUMNS_LIST_STR}[{index}].{ALLOWED_TYPES_STR}'],
                'We only support one allowed type at the moment (notice that float includes int).'
            ))
        if self.min is not None and self.max is not None:
            try:
                if self.min > self.max:
                    errors.append(ValidationRuleError(
                        [f'{COLUMNS_LIST_STR}[{index}].{MIN_STR}', f'{COLUMNS_LIST_STR}[{index}].{MAX_STR}'],
                        'Min cannot be bigger than max.'
                    ))
            except TypeError as e:
                errors.append(ValidationRuleError(
                    [f'{COLUMNS_LIST_STR}[{index}].{MIN_STR}', f'{COLUMNS_LIST_STR}[{index}].{MAX_STR}'],
                    f'Impossible to do comparison between min and max. Error: {e}'
                ))
        if self.min is not None:
            try:
                self._validate_type(self.min)
            except ValueError:
                errors.append(ValidationRuleError([f'{COLUMNS_LIST_STR}[{index}].{MIN_STR}'],
                                                  'Min value must be one of allowed types.'))
        if self.max is not None:
            try:
                self._validate_type(self.max)
            except ValueError:
                errors.append(ValidationRuleError([f'{COLUMNS_LIST_STR}[{index}].{MAX_STR}'],
                                                  'Max value must be one of allowed types.'))
        if self.allowed_types and len(self.allowed_types) == 1 and self.allowed_values:
            try:
                self._validate_type(self.allowed_values)
            except ValueError:
                errors.append(ValidationRuleError(
                    [f'{COLUMNS_LIST_STR}[{index}].{ALLOWED_VALUES_STR}',
                     f'{COLUMNS_LIST_STR}[{index}].{ALLOWED_TYPES_STR}'],
                    'All allowed values must be one of allowed type.'
                ))
        return errors

    def _validate_type(self, value):
        if not self.allowed_types:
            return
        fixed_value = [value] if not isinstance(value, list) else value
        values_df = pd.DataFrame(fixed_value)
        values_df.astype(self.allowed_types[0], errors='raise')


class TabularValidationRules:
    def __init__(self, validation_rules: dict):
        if not validation_rules:
            validation_rules = {}
        column_rules = validation_rules.get('columns', {})
        self._column_rules = column_rules
        self.preserve_order = column_rules.get(PRESERVE_ORDER_STR)
        self.column_names_required = column_rules.get(COLUMN_NAMES_REQUIRED, True)
        self.accept_other_columns = column_rules.get(ALLOW_OTHER_COLUMNS_STR, True)
        self.columns = [ColumnValidationRule(column_data) for column_data in column_rules.get('columnsList', [])]

    def validate_self(self) -> List[ValidationRuleError]:
        errors = []
        for key in self._column_rules.keys():
            if key not in [PRESERVE_ORDER_STR, COLUMN_NAMES_REQUIRED, ALLOW_OTHER_COLUMNS_STR, 'columnsList']:
                errors.append(ValidationRuleError([key], 'Invalid key'))
        if not self.column_names_required:
            errors.append(ValidationRuleError(
                [COLUMN_NAMES_REQUIRED],
                'Field must be always true. Validation by index is not implemented yet. '
            ))
        if self.accept_other_columns and self.preserve_order:
            errors.append(ValidationRuleError(
                [PRESERVE_ORDER_STR, ALLOW_OTHER_COLUMNS_STR],
                f'If you want to set preserveOrder to true, you need to explicitly set {ALLOW_OTHER_COLUMNS_STR} '
                f'to false. When we validate {PRESERVE_ORDER_STR} we get the list of all allowed columns '
                f'and compare it with the list of all columns in the file, so if there will be other columns, '
                f'the validation will fail.'
            ))

        for index, column in enumerate(self.columns):
            if self.preserve_order and not column.required:
                errors.append(ValidationRuleError(
                    [PRESERVE_ORDER_STR, f'{COLUMNS_LIST_STR}[{index}].{REQUIRED_STR}'],
                    f'If {PRESERVE_ORDER_STR} is true, then all columns must be required.'
                ))
            errors += column.validate_self(index)
        return errors


class SbioModelDataForCsv:
    # TODO make one class for csv and h5ad
    def __init__(self, model_metadata_file: BytesIO):
        reader = json.load(model_metadata_file)
        self.var_names = set(reader['columns'])
        self.target_names = set(reader['targets'])
        self.metadata = reader.get('metadata', {})



from io import BytesIO
from typing import List, Optional

import anndata
import pandas as pd

from file_process.base import FileProcessorBase
from file_process.constants import PREVIEW_ROWS_COUNT
from file_process.h5ad.h5ad_validator import H5ADValidator
from file_process.h5ad.schemas import SbioModelDataForH5ad


class H5ADFileProcessor(FileProcessorBase):

    def __init__(self, file, **_):
        self.adata = anndata.read_h5ad(file)

    def validate(self, model_metadata_file: Optional[BytesIO] = None, _: Optional[dict] = None):
        model_data = None
        if model_metadata_file:
            model_data = SbioModelDataForH5ad(model_metadata_file)
        validator = H5ADValidator(self.adata, model_data, enable_warnings=False)
        validator()

    def get_targets(self):
        return list(self.adata.obs.columns)

    def get_observations(self, rows_number: Optional[int] = None):
        return self.adata.obs.head(n=rows_number)

    def get_variables(self, rows_number: Optional[int] = None):
        return self.adata.var.head(n=rows_number)

    def get_preview(self):
        target_names = self.get_targets()
        obs_preview = self.get_observations(PREVIEW_ROWS_COUNT)
        var_preview = self.get_variables(PREVIEW_ROWS_COUNT)
        return target_names, self.create_tabular_response(obs_preview), self.create_tabular_response(var_preview)

    @staticmethod
    def create_tabular_response(data_df: pd.DataFrame) -> List[dict]:
        if data_df is None:
            return []
        data_df = data_df.astype(object)
        rows = data_df.round(2).where(pd.notnull(data_df), None).to_dict(orient='records')
        indices = list(data_df.index)
        if len(rows) != len(indices):
            return rows
        for index, value in enumerate(indices):
            rows[index]['Feature Name'] = value
        return rows



from setuptools import setup

setup(name="file_process",
      version="1.2.5",
      description="A package that does file validation and file preview.",
      license='MIT',
      author="superbio.ai",
      url='https://github.com/Superbio-ai/file-process',
      install_requires=['pandas==1.2.5', 'anndata==0.8.0', 'scanpy==1.9.1', 'numpy==1.21'],
      packages=['file_process', 'file_process.h5ad', 'file_process.csv'])



from typing import Optional

from pandas import DataFrame

from file_process.exceptions import NotAllTargetsError, NotSomeTargetsError, ModelFileValidationVariablesError, \
    CustomValidationException
from file_process.csv.schemas import TabularValidationRules, SbioModelDataForCsv, ColumnValidationRule


class CSVValidator:
    def __init__(self, data_df: DataFrame, validation_rules: Optional[dict],
                 model_data: Optional[SbioModelDataForCsv] = None):
        self.data_df = data_df
        self.rules = TabularValidationRules(validation_rules)
        self.model_data = model_data

    def __call__(self):
        self.validate()
        self.model_file_validation()

    def validate(self):
        if not self.rules:
            return
        if self.rules.column_names_required:
            self._validate_column_names()
            self._validate_columns()

    def _validate_column_names(self):
        column_names = self.data_df.columns.values.tolist()
        if self.rules.preserve_order:
            # all columns are considered as required with this filter
            required_columns = [col.name for col in self.rules.columns]
            if required_columns != column_names:
                raise CustomValidationException(f'This is the list of allowed columns in the allowed order: '
                                                f'[{required_columns}]')
        for col in self.rules.columns:
            if col.required and col.name not in column_names:
                all_required_columns = [col.name for col in self.rules.columns if col.required]
                raise CustomValidationException(f'Missing {col.name} column in the file. '
                                                f'List of required columns: [{all_required_columns}]')
        if not self.rules.accept_other_columns:
            allowed_column_names = [col.name for col in self.rules.columns]
            for col in column_names:
                if col not in allowed_column_names:
                    raise CustomValidationException(f'Invalid column: {col}. '
                                                    f'The list of allowed column names: {allowed_column_names}')

    def _validate_columns(self):
        rules = {c.name: c for c in self.rules.columns}
        for name, data in self.data_df.iteritems():
            if name in rules:
                self._validate_column(str(name), data, rules[name])

    def _validate_column(self, name: str, data, rule: ColumnValidationRule):
        if rule.allowed_types:
            # TODO add support for a list of types
            type_ = rule.allowed_types[0]
            try:
                data.astype(type_)
            except Exception as e:
                text = str(e)
                raise CustomValidationException(f'All values under {name} column must be one of the following '
                                                f'types: {rule.allowed_types}. {text.capitalize()}.') from e
        if not rule.allow_missings:
            if data.isna().sum():
                raise CustomValidationException(f'Column {name} has missings and it is not allowed.')
        if not rule.allow_duplicates:
            if data.duplicated().any():
                raise CustomValidationException(f'Column {name} has duplicates and it is not allowed.')
        if rule.min is not None:
            if data.le(rule.min).any():
                raise CustomValidationException(f'Min value in column {name} can be {rule.min}.')
        if rule.max is not None:
            if data.ge(rule.max).any():
                raise CustomValidationException(f'Max value in column {name} can be {rule.max}.')
        if rule.allowed_values:
            if not data.isin(rule.allowed_values).all():
                raise CustomValidationException(f'For {name} column the list of allowed values is '
                                                f'{rule.allowed_values}.')

    def model_file_validation(self):
        if not self.model_data:
            return

        model_target_names = self.model_data.target_names
        dataset_vars = set(self.data_df.columns)
        all_targets = self.model_data.metadata.get('require_all_targets', 'all')

        if all_targets == 'all':
            difference = model_target_names - dataset_vars
            if difference:
                raise NotAllTargetsError(difference)
        elif all_targets == 'some':
            are_targets_valid = not model_target_names or any(elem in dataset_vars for elem in model_target_names)
            if not are_targets_valid:
                raise NotSomeTargetsError(model_target_names)
        dataset_diff = dataset_vars - model_target_names
        var_names_diff = self.model_data.var_names - model_target_names
        difference = var_names_diff - dataset_diff
        if difference:
            raise ModelFileValidationVariablesError(difference)



from io import BytesIO

import pandas as pd


class SbioModelDataForH5ad:
    def __init__(self, model_metadata_file: BytesIO):
        reader = pd.read_csv(model_metadata_file, sep=',', index_col=0)
        self.var_names = set(reader.index)



from typing import List, Optional
from io import BytesIO

import pandas as pd
from numpy import number, nan
from pandas.errors import ParserError

from file_process.base import FileProcessorBase
from file_process.constants import PREVIEW_ROWS_COUNT
from file_process.csv.csv_validator import CSVValidator
from file_process.csv.schemas import SbioModelDataForCsv
from file_process.exceptions import DelimiterError


class CSVFileProcessor(FileProcessorBase):

    def __init__(self, file: BytesIO, **kwargs):
        delimiter = kwargs.get('delimiter')
        if not delimiter:
            reader = pd.read_csv(file, sep=None, iterator=True, nrows=10)
            delimiter = reader._engine.data.dialect.delimiter  # pylint: disable=protected-access
            file.seek(0)
        self.delimiter = delimiter
        try:
            self.data_df = pd.read_csv(file, sep=self.delimiter)
        except ParserError as exc:
            raise DelimiterError() from exc

    def validate(self, model_metadata_file: Optional[BytesIO] = None, validation_rules: Optional[dict] = None):
        model_data = SbioModelDataForCsv(model_metadata_file) if model_metadata_file else None
        validator = CSVValidator(self.data_df, validation_rules, model_data)
        validator()

    def get_observations(self, rows_number: int = None):
        rows_number = min(10, self.data_df.shape[0]) if rows_number else self.data_df.shape[0]
        return self.data_df.head(rows_number)

    def get_var_names(self):
        return list(self.data_df.columns)

    def get_preview(self):
        var_names = self.get_var_names()
        obs_preview = self.get_observations(PREVIEW_ROWS_COUNT)
        return var_names, self.create_tabular_response(obs_preview), None

    @staticmethod
    def create_tabular_response(data_df: pd.DataFrame) -> List[dict]:
        if data_df is None:
            return []
        numeric_columns = data_df.select_dtypes(include=number).columns
        rows = data_df.replace({nan: None})
        rows[numeric_columns] = rows[numeric_columns].round(2)
        rows = rows.to_dict(orient='records')
        return rows
